@@ -1,3 +1,8 @@
+{-# LANGUAGE DeriveGeneric #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE DeriveTraversable #-}
+{-# LANGUAGE DeriveFoldable #-}
+{-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE ExistentialQuantification #-}
@@ -6,6 +11,9 @@
 
 -- | Framework for seeing how much a function allocates.
 --
+-- WARNING: weigh is incompatible with profiling. It reports much more
+-- allocations with profiling turned on.
+--
 -- Example:
 --
 -- @
@@ -21,6 +29,8 @@
 --         count 0 = ()
 --         count a = count (a - 1)
 -- @
+--
+-- Use 'wgroup' to group sets of tests.
 
 module Weigh
   (-- * Main entry points
@@ -29,11 +39,17 @@ module Weigh
   -- * Configuration
   ,setColumns
   ,Column(..)
+  ,setFormat
+  ,Format (..)
+  ,setConfig
+  ,Config (..)
+  ,defaultConfig
   -- * Simple combinators
   ,func
   ,io
   ,value
   ,action
+  ,wgroup
   -- * Validating combinators
   ,validateAction
   ,validateFunc
@@ -44,10 +60,12 @@ module Weigh
   ,Weight(..)
   -- * Handy utilities
   ,commas
+  ,reportGroup
   -- * Internals
   ,weighDispatch
   ,weighFunc
   ,weighAction
+  ,Grouped(..)
   )
   where
 
@@ -55,10 +73,13 @@ import Control.Applicative
 import Control.Arrow
 import Control.DeepSeq
 import Control.Monad.State
-import Data.List
+import qualified Data.Foldable as Foldable
+import qualified Data.Traversable as Traversable
+import Data.Int
+import qualified Data.List as List
 import Data.List.Split
 import Data.Maybe
-import Data.Int
+import GHC.Generics
 import Prelude
 import System.Environment
 import System.Exit
@@ -77,12 +98,18 @@ data Column = Case | Allocated | GCs| Live | Check | Max
   deriving (Show, Eq, Enum)
 
 -- | Weigh configuration.
-data Config = Config {configColumns :: [Column]}
+data Config = Config
+  { configColumns :: [Column]
+  , configPrefix :: String
+  , configFormat :: !Format
+  } deriving (Show)
+
+data Format = Plain | Markdown
   deriving (Show)
 
 -- | Weigh specification monad.
 newtype Weigh a =
-  Weigh {runWeigh :: State (Config, [(String,Action)]) a}
+  Weigh {runWeigh :: State (Config, [Grouped Action]) a}
   deriving (Monad,Functor,Applicative)
 
 -- | How much a computation weighed in at.
@@ -95,53 +122,72 @@ data Weight =
          }
   deriving (Read,Show)
 
+-- | Some grouped thing.
+data Grouped a
+  = Grouped String [Grouped a]
+  | Singleton a
+  deriving (Eq, Show, Functor, Traversable.Traversable, Foldable.Foldable, Generic)
+instance NFData a => NFData (Grouped a)
+
 -- | An action to run.
 data Action =
   forall a b. (NFData a) =>
   Action {_actionRun :: !(Either (b -> IO a) (b -> a))
          ,_actionArg :: !b
+         ,actionName :: !String
          ,actionCheck :: Weight -> Maybe String}
+instance NFData Action where rnf _ = ()
 
 --------------------------------------------------------------------------------
 -- Main-runners
 
 -- | Just run the measuring and print a report. Uses 'weighResults'.
 mainWith :: Weigh a -> IO ()
-mainWith m =
-  do (results, config) <- weighResults m
-     unless (null results)
-            (do putStrLn ""
-                putStrLn (report config results))
-     case mapMaybe (\(w,r) ->
-                      do msg <- r
-                         return (w,msg))
-                   results of
-       [] -> return ()
-       errors ->
-         do putStrLn "\nCheck problems:"
-            mapM_ (\(w,r) -> putStrLn ("  " ++ weightLabel w ++ "\n    " ++ r)) errors
-            exitWith (ExitFailure (-1))
+mainWith m = do
+  (results, config) <- weighResults m
+  unless
+    (null results)
+    (do putStrLn ""
+        putStrLn (report config results))
+  case mapMaybe
+         (\(w, r) -> do
+            msg <- r
+            return (w, msg))
+         (concatMap Foldable.toList (Foldable.toList results)) of
+    [] -> return ()
+    errors -> do
+      putStrLn "\nCheck problems:"
+      mapM_
+        (\(w, r) -> putStrLn ("  " ++ weightLabel w ++ "\n    " ++ r))
+        errors
+      exitWith (ExitFailure (-1))
 
 -- | Run the measuring and return all the results, each one may have
 -- an error.
 weighResults
-  :: Weigh a -> IO ([(Weight,Maybe String)], Config)
+  :: Weigh a -> IO ([Grouped (Weight,Maybe String)], Config)
 weighResults m = do
   args <- getArgs
-  let (config, cases) =
-        execState (runWeigh m) (defaultConfig, [])
-  result <- weighDispatch args cases
+  weighEnv <- lookupEnv "WEIGH_CASE"
+  let (config, cases) = execState (runWeigh m) (defaultConfig, [])
+  result <- weighDispatch weighEnv cases
   case result of
     Nothing -> return ([], config)
     Just weights ->
       return
-        ( map
-            (\w ->
-               case lookup (weightLabel w) cases of
-                 Nothing -> (w, Nothing)
-                 Just a -> (w, actionCheck a w))
+        ( fmap
+            (fmap
+               (\w ->
+                  case glookup (weightLabel w) cases of
+                    Nothing -> (w, Nothing)
+                    Just a -> (w, actionCheck a w)))
             weights
-        , config)
+        , config
+          { configFormat =
+              if any (== "--markdown") args
+                then Markdown
+                else configFormat config
+          })
 
 --------------------------------------------------------------------------------
 -- User DSL
@@ -152,12 +198,22 @@ defaultColumns = [Case, Allocated, GCs]
 
 -- | Default config.
 defaultConfig :: Config
-defaultConfig = Config {configColumns = defaultColumns}
+defaultConfig =
+  Config
+  {configColumns = defaultColumns, configPrefix = "", configFormat = Plain}
 
--- | Set the config. Default is: 'defaultConfig'.
+-- | Set the columns to display in the config
 setColumns :: [Column] -> Weigh ()
 setColumns cs = Weigh (modify (first (\c -> c {configColumns = cs})))
 
+-- | Set the output format in the config
+setFormat :: Format -> Weigh ()
+setFormat fm = Weigh (modify (first (\c -> c {configFormat = fm})))
+
+-- | Set the config. Default is: 'defaultConfig'.
+setConfig :: Config -> Weigh ()
+setConfig = Weigh . modify . first . const
+
 -- | Weigh a function applied to an argument.
 --
 -- Implemented in terms of 'validateFunc'.
@@ -214,7 +270,7 @@ validateAction :: (NFData a)
                -> (Weight -> Maybe String) -- ^ A validating function, returns maybe an error.
                -> Weigh ()
 validateAction name !m !arg !validate =
-  tellAction [(name,Action (Left m) arg validate)]
+  tellAction name $ flip (Action (Left m) arg) validate
 
 -- | Weigh a function, validating the result
 validateFunc :: (NFData a)
@@ -224,77 +280,96 @@ validateFunc :: (NFData a)
              -> (Weight -> Maybe String) -- ^ A validating function, returns maybe an error.
              -> Weigh ()
 validateFunc name !f !x !validate =
-  tellAction [(name,Action (Right f) x validate)]
+  tellAction name $ flip (Action (Right f) x)  validate
 
 -- | Write out an action.
-tellAction :: [(String, Action)] -> Weigh ()
-tellAction x = Weigh (modify (second ( ++ x)))
+tellAction :: String -> (String -> Action) -> Weigh ()
+tellAction name act =
+  Weigh (do prefix <- gets (configPrefix . fst)
+            modify (second (\x -> x ++ [Singleton $ act (prefix ++ "/" ++ name)])))
+
+-- | Make a grouping of tests.
+wgroup :: String -> Weigh () -> Weigh ()
+wgroup str wei = do
+  (orig, start) <- Weigh get
+  let startL = length $ start
+  Weigh (modify (first (\c -> c {configPrefix = configPrefix orig ++ "/" ++ str})))
+  wei
+  Weigh $ do
+    modify $ second $ \x -> take startL x ++ [Grouped str $ drop startL x]
+    modify (first (\c -> c {configPrefix = configPrefix orig}))
 
 --------------------------------------------------------------------------------
 -- Internal measuring actions
 
 -- | Weigh a set of actions. The value of the actions are forced
 -- completely to ensure they are fully allocated.
-weighDispatch :: [String] -- ^ Program arguments.
-              -> [(String,Action)] -- ^ Weigh name:action mapping.
-              -> IO (Maybe [Weight])
+weighDispatch :: Maybe String -- ^ The content of then env variable WEIGH_CASE.
+              -> [Grouped Action] -- ^ Weigh name:action mapping.
+              -> IO (Maybe [(Grouped Weight)])
 weighDispatch args cases =
   case args of
-    ("--case":label:fp:_) ->
+    Just var -> do
+      let (label:fp:_) = read var
       let !_ = force fp
-      in case lookup label (deepseq (map fst cases) cases) of
-           Nothing -> error "No such case!"
-           Just act -> do
-             case act of
-               Action !run arg _ -> do
-                 (bytes, gcs, liveBytes, maxByte) <-
-                   case run of
-                     Right f -> weighFunc f arg
-                     Left m -> weighAction m arg
-                 writeFile
-                   fp
-                   (show
-                      (Weight
-                       { weightLabel = label
-                       , weightAllocatedBytes = bytes
-                       , weightGCs = gcs
-                       , weightLiveBytes = liveBytes
-                       , weightMaxBytes = maxByte
-                       }))
-             return Nothing
-    _
-      | names == nub names -> fmap Just (mapM (fork . fst) cases)
-      | otherwise -> error "Non-unique names specified for things to measure."
-      where names = map fst cases
+      case glookup label (force cases) of
+        Nothing -> error "No such case!"
+        Just act -> do
+          case act of
+            Action !run arg _ _ -> do
+              (bytes, gcs, liveBytes, maxByte) <-
+                case run of
+                  Right f -> weighFunc f arg
+                  Left m -> weighAction m arg
+              writeFile
+                fp
+                (show
+                   (Weight
+                    { weightLabel = label
+                    , weightAllocatedBytes = bytes
+                    , weightGCs = gcs
+                    , weightLiveBytes = liveBytes
+                    , weightMaxBytes = maxByte
+                    }))
+          return Nothing
+    _ -> fmap Just (Traversable.traverse (Traversable.traverse fork) cases)
+
+-- | Lookup an action.
+glookup :: String -> [Grouped Action] -> Maybe Action
+glookup label =
+  Foldable.find ((== label) . actionName) .
+  concat . map Foldable.toList . Foldable.toList
 
 -- | Fork a case and run it.
-fork :: String -- ^ Label for the case.
+fork :: Action -- ^ Label for the case.
      -> IO Weight
-fork label =
+fork act =
   withSystemTempFile
     "weigh"
     (\fp h -> do
        hClose h
+       setEnv "WEIGH_CASE" $ show $ [actionName act,fp]
        me <- getExecutablePath
        (exit, _, err) <-
          readProcessWithExitCode
            me
-           ["--case", label, fp, "+RTS", "-T", "-RTS"]
+           ["+RTS", "-T", "-RTS"]
            ""
        case exit of
          ExitFailure {} ->
-           error ("Error in case (" ++ show label ++ "):\n  " ++ err)
-         ExitSuccess ->
-           do out <- readFile fp
-              case reads out of
-                [(!r, _)] -> return r
-                _ ->
-                  error
-                    (concat
-                       [ "Malformed output from subprocess. Weigh"
-                       , " (currently) communicates with its sub-"
-                       , "processes via a temporary file."
-                       ]))
+           error
+             ("Error in case (" ++ show (actionName act) ++ "):\n  " ++ err)
+         ExitSuccess -> do
+           out <- readFile fp
+           case reads out of
+             [(!r, _)] -> return r
+             _ ->
+               error
+                 (concat
+                    [ "Malformed output from subprocess. Weigh"
+                    , " (currently) communicates with its sub-"
+                    , "processes via a temporary file."
+                    ]))
 
 -- | Weigh a pure function. This function is heavily documented inside.
 weighFunc
@@ -341,7 +416,7 @@ weighFunc run !arg = do
     , fromIntegral liveBytes
     , fromIntegral maxBytes)
 
--- | Weigh a pure function. This function is heavily documented inside.
+-- | Weigh an IO action. This function is heavily documented inside.
 weighAction
   :: (NFData a)
   => (b -> IO a)      -- ^ A function whose memory use we want to measure.
@@ -389,10 +464,46 @@ weighAction run !arg = do
 --------------------------------------------------------------------------------
 -- Formatting functions
 
--- | Make a report of the weights.
-report :: Config -> [(Weight,Maybe String)] -> String
-report config = tablize . (select headings :) . map (select . toRow)
+report :: Config -> [Grouped (Weight,Maybe String)] -> String
+report config gs =
+  List.intercalate
+    "\n\n"
+    (filter
+       (not . null)
+       [ if null singletons
+           then []
+           else reportTabular config singletons
+       , List.intercalate "\n\n" (map (uncurry (reportGroup config)) groups)
+       ])
   where
+    singletons =
+      mapMaybe
+        (\case
+           Singleton v -> Just v
+           _ -> Nothing)
+        gs
+    groups =
+      mapMaybe
+        (\case
+           Grouped title vs -> Just (title, vs)
+           _ -> Nothing)
+        gs
+
+reportGroup :: Config -> [Char] -> [Grouped (Weight, Maybe String)] -> [Char]
+reportGroup config title gs =
+  case configFormat config of
+    Plain -> title ++ "\n\n" ++ indent (report config gs)
+    Markdown -> "#" ++ title ++ "\n\n" ++ report config gs
+
+-- | Make a report of the weights.
+reportTabular :: Config -> [(Weight,Maybe String)] -> String
+reportTabular config = tabled
+  where
+    tabled =
+      (case configFormat config of
+         Plain -> tablize
+         Markdown -> mdtable) .
+      (select headings :) . map (select . toRow)
     select row = mapMaybe (\name -> lookup name row) (configColumns config)
     headings =
       [ (Case, (True, "Case"))
@@ -403,7 +514,7 @@ report config = tablize . (select headings :) . map (select . toRow)
       , (Max, (False, "Max"))
       ]
     toRow (w, err) =
-      [ (Case, (True, weightLabel w))
+      [ (Case, (True, takeLastAfterBk $ weightLabel w))
       , (Allocated, (False, commas (weightAllocatedBytes w)))
       , (GCs, (False, commas (weightGCs w)))
       , (Live, (False, commas (weightLiveBytes w)))
@@ -414,18 +525,45 @@ report config = tablize . (select headings :) . map (select . toRow)
               Nothing -> "OK"
               Just {} -> "INVALID"))
       ]
+    takeLastAfterBk w = case List.elemIndices '/' w of
+                       [] -> w
+                       x  -> drop (1+last x) w
+
+-- | Make a markdown table.
+mdtable ::[[(Bool,String)]] -> String
+mdtable rows = List.intercalate "\n" [heading, align, body]
+  where
+    heading = columns (map (\(_, str) -> str) (fromMaybe [] (listToMaybe rows)))
+    align =
+      columns
+        (map
+           (\(shouldAlignLeft, _) ->
+              if shouldAlignLeft
+                then ":---"
+                else "---:")
+           (fromMaybe [] (listToMaybe rows)))
+    body =
+      List.intercalate "\n" (map (\row -> columns (map snd row)) (drop 1 rows))
+    columns xs = "|" ++ List.intercalate "|" xs ++ "|"
 
 -- | Make a table out of a list of rows.
 tablize :: [[(Bool,String)]] -> String
 tablize xs =
-  intercalate "\n"
-              (map (intercalate "  " . map fill . zip [0 ..]) xs)
-  where fill (x',(left',text')) = printf ("%" ++ direction ++ show width ++ "s") text'
-          where direction = if left'
-                               then "-"
-                               else ""
-                width = maximum (map (length . snd . (!! x')) xs)
+  List.intercalate "\n" (map (List.intercalate "  " . map fill . zip [0 ..]) xs)
+  where
+    fill (x', (left', text')) =
+      printf ("%" ++ direction ++ show width ++ "s") text'
+      where
+        direction =
+          if left'
+            then "-"
+            else ""
+        width = maximum (map (length . snd . (!! x')) xs)
 
 -- | Formatting an integral number to 1,000,000, etc.
 commas :: (Num a,Integral a,Show a) => a -> String
-commas = reverse . intercalate "," . chunksOf 3 . reverse . show
+commas = reverse . List.intercalate "," . chunksOf 3 . reverse . show
+
+-- | Indent all lines in a string.
+indent :: [Char] -> [Char]
+indent = List.intercalate "\n" . map (replicate 2 ' '++) . lines
